#include<ros/ros.h>
#include<pcl_ros/point_cloud.h>
#include<pcl_conversions/pcl_conversions.h>
#include<sensor_msgs/PointCloud2.h>

#include <pcl/visualization/cloud_viewer.h>
#include <iostream>
#include <pcl/io/pcd_io.h>
#include<pcl/point_cloud.h>
#include <pcl/point_types.h>
#include <pcl/filters/filter.h>
#include <pcl/filters/voxel_grid.h>
#include <pcl/filters/passthrough.h>

#include <PCLFilters.h>

ros::Publisher pub;

PassThroughFilter pass;
VoxelGridFilter down;


pcl::visualization::CloudViewer viewer("Cloud Viewer");

int count=0;
char filename[32];

void  pass_through (pcl::PointCloud<pcl::PointXYZ>::Ptr cloud)
{
    pcl::PointCloud<pcl::PointXYZ>::Ptr cloud_filtered_y (new pcl::PointCloud<pcl::PointXYZ>);
    pcl::PointCloud<pcl::PointXYZ>::Ptr cloud_filtered_z (new pcl::PointCloud<pcl::PointXYZ>);

    double min_x;
    double max_x;
    double min_y;
    double max_y;
    double min_z;
    double max_z;
    double right;
    ros::param::get("~MIN_X",min_x);
    ros::param::get("~MAX_X",max_x);
    ros::param::get("~MIN_Y",min_y);
    ros::param::get("~MAX_Y",max_y);
    ros::param::get("~MIN_Z",min_z);
    ros::param::get("~MAX_Z",max_z);
    ros::param::get("~Right_Cut",right);

    pass.filter(cloud,cloud_filtered_y,'y',min_y,max_y);
    pass.filter(cloud_filtered_y,cloud_filtered_z,'z',min_z,max_z);
    pass.filter(cloud_filtered_z,cloud,'x',min_x,max_x, true);
    pass.filter(cloud,cloud,'x',right,2, true);

}

void
cloud_cb (const sensor_msgs::PointCloud2ConstPtr& cloud_msg)
{
    pcl::PointCloud<pcl::PointXYZ>::Ptr cloud(new pcl::PointCloud<pcl::PointXYZ>);
    pcl::PointCloud<pcl::PointXYZ>::Ptr cloud_remove_nan(new pcl::PointCloud<pcl::PointXYZ>);
    pcl::PointCloud<pcl::PointXYZ>::Ptr cloud_down(new pcl::PointCloud<pcl::PointXYZ>);

    pcl::fromROSMsg (*cloud_msg, *cloud);

    //remove nan
    RemoveNaN::removeNaN(*cloud,*cloud_remove_nan);

    bool pass_flag;
    ros::param::get("~pass_flag",pass_flag);
    if(pass_flag==true)
    {
        void  pass_through (pcl::PointCloud<pcl::PointXYZ>::Ptr );
        pass_through(cloud_remove_nan);

    }





    bool down_flag;
    ros::param::get("~down_flag",down_flag);
    if(down_flag==true)
    {
        down.filter(cloud_remove_nan,cloud_remove_nan);
    }

    viewer.showCloud(cloud_remove_nan);



       //save point cloud full or down
    bool save_flag;
    ros::param::get("~save_flag",save_flag);
    if(save_flag==true)
    {
      sprintf(filename,"%d.pcd",count);
      pcl::io::savePCDFileASCII (filename, *cloud_remove_nan);
      std::cout << "save a picture, named "<< count<<".pcd"<<std::endl;
      count++;
      ros::param::set("~save_flag",false);
    }




    sensor_msgs::PointCloud2 output;
    pcl::toROSMsg (*cloud_remove_nan, output);
    pub.publish (output);
}

int
main (int argc, char** argv)
{
  // Initialize ROS
  ros::init (argc, argv, "viewer");
  ros::NodeHandle nh;

  // Create a ROS subscriber for the input point cloud
  ros::Subscriber sub = nh.subscribe<sensor_msgs::PointCloud2> ("/camera/depth/points", 1, cloud_cb);

  pub = nh.advertise<sensor_msgs::PointCloud2> ("cloud_remove_nan", 1);

  // Spin
  ros::spin ();
}


