#include "../include/PCLRegistration.h"

void SAC_IA::SAC_IA_Registration(pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud_src,pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud_tgt, double radius_normal, double radius_fpfh)
{
    pcl::NormalEstimation<pcl::PointXYZ, pcl::Normal> ne;
    pcl::FPFHEstimation<pcl::PointXYZ, pcl::Normal, pcl::FPFHSignature33> fpfh;
    //src
    // Estimate the normals.
    ne.setInputCloud(cloud_src);
    pcl::search::KdTree<pcl::PointXYZ>::Ptr kdtree_src(new pcl::search::KdTree<pcl::PointXYZ>);
    ne.setSearchMethod(kdtree_src);
    ne.setRadiusSearch(radius_normal);
    pcl::PointCloud<pcl::Normal>::Ptr normals_src (new pcl::PointCloud<pcl::Normal>);
    ne.compute(*normals_src);

    pcl::PointCloud<pcl::FPFHSignature33>::Ptr fpfh_src (new pcl::PointCloud<pcl::FPFHSignature33>);
    // FPFH estimation object.
    fpfh.setInputCloud(cloud_src);
    fpfh.setInputNormals(normals_src);
    pcl::search::KdTree<pcl::PointXYZ>::Ptr tree_src (new pcl::search::KdTree<pcl::PointXYZ> () );
    fpfh.setSearchMethod(tree_src);
    // using the total number of points.
    fpfh.setRadiusSearch (radius_fpfh);
    fpfh.compute(*fpfh_src);

    //tgt
    // Estimate the normals.
    ne.setInputCloud(cloud_tgt);
    pcl::search::KdTree<pcl::PointXYZ>::Ptr kdtree_tgt(new pcl::search::KdTree<pcl::PointXYZ>);
    ne.setSearchMethod(kdtree_tgt);
    ne.setRadiusSearch(radius_normal);
    pcl::PointCloud<pcl::Normal>::Ptr normals_tgt (new pcl::PointCloud<pcl::Normal>);
    ne.compute(*normals_tgt);

    pcl::PointCloud<pcl::FPFHSignature33>::Ptr fpfh_tgt (new pcl::PointCloud<pcl::FPFHSignature33>);
    // FPFH estimation object.
    fpfh.setInputCloud(cloud_tgt);
    fpfh.setInputNormals(normals_tgt);
    pcl::search::KdTree<pcl::PointXYZ>::Ptr tree_tgt (new pcl::search::KdTree<pcl::PointXYZ> () );
    fpfh.setSearchMethod(tree_tgt);
    // using the total number of points.
    fpfh.setRadiusSearch (radius_fpfh);
    fpfh.compute(*fpfh_tgt);

    pcl::SampleConsensusInitialAlignment<pcl::PointXYZ, pcl::PointXYZ, pcl::FPFHSignature33> scia;
    scia.setInputSource(cloud_src);
    scia.setInputTarget(cloud_tgt);
    scia.setSourceFeatures(fpfh_src);
    scia.setTargetFeatures(fpfh_tgt);
    //scia.setMinSampleDistance(1);
    //scia.setNumberOfSamples(2);
    //scia.setCorrespondenceRandomness(20);
    cloud_sac_.reset (new pcl::PointCloud<pcl::PointXYZ>);
    scia.align(*cloud_sac_);
    std::cout  <<"sac has converged:"<<scia.hasConverged()<<"  score: "<<scia.getFitnessScore()<<std::endl;
    score_ = scia.getFitnessScore();
    sac_trans_=scia.getFinalTransformation().cast<double>();
    std::cout<<sac_trans_<<std::endl;
    pcl::io::savePCDFileASCII("saced.pcd",*cloud_sac_);
}



void ICP::ICP_Registration(pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud_src,pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud_tgt,  Eigen::Matrix4d transform_init, double MaxCorrespondenceDistance, int MaximumIterations, double TransformationEpsilon, double EuclideanFitnessEpsilon  )
{
    pcl::IterativeClosestPoint<pcl::PointXYZ, pcl::PointXYZ> icp;
    icp.setInputSource(cloud_src);
    icp.setInputTarget(cloud_tgt);
    icp.setMaxCorrespondenceDistance(MaxCorrespondenceDistance);
    icp.setTransformationEpsilon(TransformationEpsilon);
    icp.setEuclideanFitnessEpsilon(EuclideanFitnessEpsilon);
    icp.setMaximumIterations(MaximumIterations);
    cloud_icp_.reset (new pcl::PointCloud<pcl::PointXYZ>);
    icp.align(*cloud_icp_,transform_init.cast<float>());


    std::cout << "icp has converged:" << icp.hasConverged() << " score: " <<icp.getFitnessScore() << std::endl;
    score_ = icp.getFitnessScore();
    icp_trans_=icp.getFinalTransformation().cast<double>();
    std::cout << icp_trans_ << std::endl;

    pcl::io::savePCDFileASCII ("icped.pcd", *cloud_icp_);

}

void TRANS_ICP::trans_icp(pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud_src,pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud_tgt, double MaxCorrespondenceDistance, int MaximumIterations, double TransformationEpsilon, double EuclideanFitnessEpsilon)
{
    static double last = pcl::getTime ();

    Eigen::Vector4d centroid_src;
    pcl::compute3DCentroid(*cloud_src, centroid_src);

    Eigen::Vector4d centroid_tgt;
    pcl::compute3DCentroid(*cloud_tgt, centroid_tgt);

    Eigen::Vector4d translation;
    translation=centroid_tgt-centroid_src;
    translation[3]=1;

    Eigen::Matrix4d transform_init;
    transform_init.col(0)<<1,0,0,0;
    transform_init.col(1)<<0,1,0,0;
    transform_init.col(2)<<0,0,1,0;
    transform_init.col(3)=translation;
    std::cout<<"transform_init:"<<std::endl<<transform_init<<std::endl;
    trans_=transform_init;

    ICP icp;
    icp.ICP_Registration(cloud_src,cloud_tgt,transform_init,MaxCorrespondenceDistance,MaximumIterations,TransformationEpsilon,EuclideanFitnessEpsilon);
    cloud_transicp_=icp.cloud_icp_;
    transicp_trans_=icp.icp_trans_;//has included transform_init
    score_ = icp.score_;

    double now = pcl::getTime ();
    std::cout << "Trans_icp Time:"<<double(now - last) << " s" <<  std::endl;
}

void SAC_ICP::sac_icp(pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud_src,pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud_tgt, double radius_normal, double radius_fpfh, double MaxCorrespondenceDistance, int MaximumIterations, double TransformationEpsilon, double EuclideanFitnessEpsilon)
{
      static double last = pcl::getTime ();

      SAC_IA sac;
      sac.SAC_IA_Registration(cloud_src,cloud_tgt,radius_normal,radius_fpfh);
      sac_=sac.sac_trans_;

      ICP icp;
      icp.ICP_Registration(cloud_src,cloud_tgt,sac_,MaxCorrespondenceDistance,MaximumIterations,TransformationEpsilon,EuclideanFitnessEpsilon);
      sacicp_trans_=icp.icp_trans_;
      cloud_sacicp_=icp.cloud_icp_;
      score_ = icp.score_;


      double now = pcl::getTime ();
      std::cout << double(now - last) << " s" <<  std::endl;
}
