#include <ros/ros.h>
#include <string.h>
#include "boost/units/io.hpp"	//must---
#include <actionlib/client/simple_action_client.h>
#include <ik/ikAction.h>
#include <geometry_msgs/Pose.h>
#include<geometry_msgs/PoseStamped.h>
#include "brics_actuator/JointPositions.h"
#include "geometry_msgs/Twist.h"
#include <std_msgs/Int32.h>
#include "arm_interface_youbot.h"
#include <Eigen/Eigen>

//#define DEBUG

using namespace std;
std_msgs::Int32 controllerState;   
ros::Publisher platformPublisher;
ros::Subscriber Marker_Pos_Subscriber;
ros::Publisher statePub;
ros::Publisher armPublisher;

cArmInterfaceYoubot* pArmInterfaceYoubot;

typedef actionlib::SimpleActionClient<ik::ikAction> Client;

std::vector<double> grapper_point_jointvalues(5);	//存储“目标点”轴值
std::vector<double> pre_jointvalues(5);	//存储“预抓点”轴值
bool grab_ik = false;  //计算“目标点”是否逆解，定义标志位
bool pre_ik = false;  //计算“预抓点”是否逆解，定义标志位
void grasp_action();
void Arm_Place_Front();
void Arm_Ready();



void mark_pose_callback(const geometry_msgs::PoseStamped &msg)
{
    ROS_INFO("IK solver is comein");

    geometry_msgs::Pose obj_pose;
    geometry_msgs::Pose grab_pose;	//定义“目标点”位姿
    geometry_msgs::Pose pre_pose;	//定义“预抓点”位姿
    obj_pose = msg.pose;  //暂存接收到的mark坐标值

    grab_ik = false;    //计算“目标点”是否逆解，标志位false
    pre_ik = false;    //计算“目标点”是否逆解，标志位false

    Eigen::Vector3f t_grab(-0.0175,-0,0.05);
    Eigen::Vector3f t_pre(-0.0175,-0,0.10);

    Eigen::Matrix3f R_obj,R;
    Eigen::Quaternionf tmp;
    tmp.x()=obj_pose.orientation.x;
    tmp.y()=obj_pose.orientation.y;
    tmp.z()=obj_pose.orientation.z;
    tmp.w()=obj_pose.orientation.w;
    R_obj = tmp.toRotationMatrix();
#ifdef DEBUG
    std::cout<<R_obj<<std::endl;
#endif

    t_grab = R_obj*t_grab;
    t_pre = R_obj*t_pre;

    Eigen::Vector3f x_axis, y_axis, z_axis;
    x_axis = -R_obj.col(1); //extract the x axis
    z_axis = -R_obj.col(2); //define gripper z axis antiparallel to object z-axis
    y_axis = z_axis.cross(x_axis); // construct a right-hand coordinate frame
    R.col(0) = x_axis;
    R.col(1) = y_axis;
    R.col(2) = z_axis;
    Eigen::Quaternionf q (R);


    grab_pose.position.x=obj_pose.position.x+t_grab[0];
    grab_pose.position.y=obj_pose.position.y+t_grab[1];
    grab_pose.position.z=obj_pose.position.z+t_grab[2];
    grab_pose.orientation.x = q.x();
    grab_pose.orientation.y = q.y();
    grab_pose.orientation.z = q.z();
    grab_pose.orientation.w = q.w();


    pre_pose.position.x=obj_pose.position.x+t_pre[0];
    pre_pose.position.y=obj_pose.position.y+t_pre[1];
    pre_pose.position.z=obj_pose.position.z+t_pre[2];
    pre_pose.orientation.x = q.x();
    pre_pose.orientation.y = q.y();
    pre_pose.orientation.z = q.z();
    pre_pose.orientation.w = q.w();


    Client ik_client("youbot_ik",true);
    ik_client.waitForServer();
    ik::ikGoal goal_pose;
//---第一次计算---------------------------------------------------
    goal_pose.pose=grab_pose;
    ik_client.sendGoal(goal_pose);
    ik_client.waitForResult(ros::Duration(10.0));

    if(ik_client.getState()==actionlib::SimpleClientGoalState::SUCCEEDED)   //getState()
    {
        //如果逆解成功，则“抓取系列动作”使能
        ROS_INFO("IK1 solver is ok");
   		ik::ikResult p=*(ik_client.getResult());

        //ROS_INFO("have %d joints",p.jointresult.size());

        std::cout<<"have "<<p.jointresult.size()<<" joints"<<std::endl;
   		for(int i=0;i<p.jointresult.size();i++)
   		{
        	std::cout<<p.jointresult[i]<<std::endl;
                grapper_point_jointvalues[i]=p.jointresult[i];
            //这里需要把joints值，通过变量传出来
            //ROS_INFO("id:%f",msg.markers[0].id);
   		}
        grab_ik = true;
		//	pArmInterfaceYoubot->PublishJointValues(jointvalues);
		//	pArmInterfaceYoubot->PublishGripperValues(0.025);     
		//	moveArm_cont(jointvalues);
		//	std::cout<<"get the ik valve"<<std::endl;
		//使能标志位，手臂抓取动作函数
    }
    else
    {
        grab_ik = false;
        ROS_INFO("IK1 solver is false");
    }
//--第二次计算----------------------------------------------
    goal_pose.pose=pre_pose;
    ik_client.sendGoal(goal_pose);
    ik_client.waitForResult(ros::Duration(10.0));

    if(ik_client.getState()==actionlib::SimpleClientGoalState::SUCCEEDED)   //getState()
    {
        //如果逆解成功，则“抓取系列动作”使能
        ROS_INFO("IK2 solver is ok");
   		ik::ikResult p=*(ik_client.getResult());

        //ROS_INFO("have %d joints",p.jointresult.size());

    	std::cout<<"have "<<p.jointresult.size()<<" joints"<<std::endl;
 	  	for(int i=0;i<p.jointresult.size();i++)
 	  	{
 	       	std::cout<<p.jointresult[i]<<std::endl;
                pre_jointvalues[i]=p.jointresult[i];
 	       	//这里需要把joints值，通过变量传出来
           	//ROS_INFO("id:%f",msg.markers[0].id);
   		}
                pre_ik = true;
    }
    else
    {
        pre_ik = false;
        ROS_INFO("IK2 solver is false");
    }
	grasp_action();
//---------------------------------------------------------
    ros::Duration(5).sleep();
}



void grasp_action()
{
    if(grab_ik && pre_ik)
    {
        ROS_INFO("ARM Ik is ok");
//	pArmInterfaceYoubot -> Arm_Grasp(grapper_point_jointvalues,pre_jointvalues);
        pArmInterfaceYoubot->PublishJointValues(pre_jointvalues);
        ros::Duration(1).sleep();
        pArmInterfaceYoubot->OpenGrippers();
        ros::Duration(3).sleep();
        pArmInterfaceYoubot->PublishJointValues(grapper_point_jointvalues);
        ros::Duration(2).sleep();
        pArmInterfaceYoubot->CloseGrippers();
        ros::Duration(3).sleep();
        pArmInterfaceYoubot->PublishJointValues(pre_jointvalues);
        ros::Duration(2).sleep();

        Arm_Place_Front();
        ros::Duration(2).sleep();
        pArmInterfaceYoubot->OpenGrippers();
        ros::Duration(3).sleep();
        Arm_Ready();
    }
	else
	{
        controllerState.data = 1;
        statePub.publish(controllerState);
	}
}

void Arm_Place_Front()
{
    std::vector<double> jointvalues(5);

    // move arm high position. values were determined empirically
    jointvalues[0] = 2.96705;		//A1  0.0100692-5.84014
    jointvalues[1] = 1.9448;		//A2  0.0100692-2.61799
    jointvalues[2] = -1.50367;		//A3  -5.02655-0.015708
    jointvalues[3] = 3.07571;		//A4  0.0221239-3.4292
    jointvalues[4] = 2.92343;		//A5  0.110619-5.64159

    int numberOfJoints = 5;
    brics_actuator::JointPositions pos_high;

//	if (newPositions.size() < numberOfJoints)
//		return 0;
//                break;
    for( int i = 0; i < numberOfJoints; ++i )
    {
            brics_actuator::JointValue val;
            val.timeStamp = ros::Time::now();
            val.value = jointvalues[i];
            val.unit = "rad";
            std::ostringstream jointName;
            jointName << "arm_joint_" << (i+1);
            val.joint_uri = jointName.str();
            pos_high.positions.push_back(val);
    }
    armPublisher.publish(pos_high);
    ros::Duration(2).sleep();




        // move arm low position. values were determined empirically
        jointvalues[0] =  2.96705;//2.96706;//-0.04  //2.96705;//-0.05    //2.96707;(-0.02)	//A1  0.0100692-5.84014
        jointvalues[1] =  2.21209;//2.16936;         //2.21209;             //2.091;		//A2  0.0100692-2.61799
        jointvalues[2] =  -1.37979; //-1.37671;        //-1.37979;            //-1.38589;		//A3  -5.02655-0.015708
        jointvalues[3] =  2.68454; //2.72419;         //2.68454;             //2.81173;		//A4  0.0221239-3.4292
        jointvalues[4] =  2.92342;//2.92342;         //2.92342;             //2.92343;		//A5  0.110619-5.64159

        brics_actuator::JointPositions pos_low;

//	if (newPositions.size() < numberOfJoints)
//		return 0;
//                break;
        for( int i = 0; i < numberOfJoints; ++i )
        {
                brics_actuator::JointValue val;
                val.timeStamp = ros::Time::now();
                val.value = jointvalues[i];
                val.unit = "rad";
                std::ostringstream jointName;
                jointName << "arm_joint_" << (i+1);
                val.joint_uri = jointName.str();
                pos_low.positions.push_back(val);
        }
        armPublisher.publish(pos_low);
}

void Arm_Ready()
{
        std::vector<double> jointvalues(5);

        // move arm straight up. values were determined empirically
        jointvalues[0] = 2.96705;		//A1  0.0100692-5.84014
        jointvalues[1] = 1.88132;		//A2  0.0100692-2.61799
        jointvalues[2] = -1.53989;		//A3  -5.02655-0.015708
        jointvalues[3] = 3.17541;		//A4  0.0221239-3.4292
        jointvalues[4] = 2.92341;		//A5  0.110619-5.64159

        int numberOfJoints = 5;
        brics_actuator::JointPositions pos;

//	if (newPositions.size() < numberOfJoints)
//		return 0;
//                break;
        for( int i = 0; i < numberOfJoints; ++i )
        {
                brics_actuator::JointValue val;
                val.timeStamp = ros::Time::now();
                val.value = jointvalues[i];
                val.unit = "rad";
                std::ostringstream jointName;
                jointName << "arm_joint_" << (i+1);
                val.joint_uri = jointName.str();
                pos.positions.push_back(val);
        }
        armPublisher.publish(pos);
}



int main(int argc,char **argv)
{
        ros::init(argc, argv, "youbot_ik_pick_client");
	ros::NodeHandle n;
        ros::NodeHandle ph("~");
	pArmInterfaceYoubot= new cArmInterfaceYoubot(ph);

        Marker_Pos_Subscriber = n.subscribe("pose",1,&mark_pose_callback);
        armPublisher = n.advertise<brics_actuator::JointPositions>("arm_1/arm_controller/position_command", 1);
//	platformPublisher = n.advertise<geometry_msgs::Twist>("cmd_vel", 1);
   	statePub=n.advertise<std_msgs::Int32>("/control_current_state", 1);
	sleep(1);	//must!!!!!  must!!!!!


	pArmInterfaceYoubot -> Arm_Upright();
	ros::Duration(4).sleep();
        //pArmInterfaceYoubot -> Arm_Recognition_Ready();
        //ros::Duration(4).sleep();
	controllerState.data = 1;
        statePub.publish(controllerState);

	sleep(1);
	ros::spin();
}






